/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

package ch.epfl.lia.frodo.algorithms.adopt.test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import junit.extensions.RepeatedTest;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.jdom.Element;

import ch.epfl.lia.frodo.algorithms.AgentInterface;
import ch.epfl.lia.frodo.algorithms.RandGraphFactory;
import ch.epfl.lia.frodo.algorithms.XCSPparser;
import ch.epfl.lia.frodo.algorithms.RandGraphFactory.Graph;
import ch.epfl.lia.frodo.algorithms.adopt.BoundsMsg;
import ch.epfl.lia.frodo.algorithms.adopt.Preprocessing;
import ch.epfl.lia.frodo.algorithms.test.AllTests;
import ch.epfl.lia.frodo.algorithms.varOrdering.dfs.DFSgeneration.DFSview;
import ch.epfl.lia.frodo.algorithms.varOrdering.dfs.DFSgeneration.MessageDFSoutput;
import ch.epfl.lia.frodo.communication.IncomingMsgPolicyInterface;
import ch.epfl.lia.frodo.communication.Message;
import ch.epfl.lia.frodo.communication.MessageWrapper;
import ch.epfl.lia.frodo.communication.Queue;
import ch.epfl.lia.frodo.communication.sharedMemory.QueueIOPipe;
import ch.epfl.lia.frodo.solutionSpaces.AddableInteger;
import ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.hypercube.Hypercube;

/** JUnit tests for the Preprocessing module
 * @author Brammert Ottens
 */
public class testPreprocessing extends TestCase {

	/** Maximum number of variables in the random graph 
	 * @note Must be at least 2. 
	 */
	private final int maxNbrVars = 5;
	
	/** Maximum number of edges in the random graph */
	private final int maxNbrEdges = 10;

	/** Maximum number of agents */
	private final int maxNbrAgents = 5;
	
	/** The queue we are listening to*/
	private Queue[] queues;
	
	/** Used to make the test thread wait */
	private final ReentrantLock finished_lock = new ReentrantLock ();
	
	/** Used to wake up the test thread when all agents have finished */
	private final Condition finished = finished_lock.newCondition();

	/** Counts the number of messages received*/
	private int counter;
	
	/** A list variables*/
	ArrayList<String> variables;
	
	/** The domain of all the variables*/
	HashMap<String, AddableInteger[]> domains;
	
	/** The random problem */
	private XCSPparser<AddableInteger, AddableInteger> parser;
	
	/** The DFS belonging to the problem */
	private Map<String, DFSview> dfs;
	
	/** The constraint graph of the problem */
	private Graph graph;
	
	/** The proper h values, calculated centrally */
	private HashMap<String, UtilitySolutionSpace<AddableInteger, AddableInteger>> precalculated_h;
	
	/** Whether to count NCCCs */
	private boolean countNCCCs = false;
	
	/** Constructor that instantiates a test only for the input method
	 * @param method test method
	 */
	public testPreprocessing(String method) {
		super (method);
	}
	
	/** @return the test suite for this test */
	public static TestSuite suite () {
		TestSuite testSuite = new TestSuite ("Tests for Preprocessing");
		
		TestSuite testTmp = new TestSuite ("Tests for Simple Heuristic without NCCCs");
		testTmp.addTest(new RepeatedTest (new testPreprocessing ("testSimpleHeuristicNoNCCCs"), 10));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for Simple Heuristic with NCCCs");
		testTmp.addTest(new RepeatedTest (new testPreprocessing ("testSimpleHeuristicWithNCCCs"), 10));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for DP0 heuristic without NCCCs");
		testTmp.addTest(new RepeatedTest (new testPreprocessing ("testDP0HeuristicNoNCCCs"), 500));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for DP0 heuristic with NCCCs");
		testTmp.addTest(new RepeatedTest (new testPreprocessing ("testDP0HeuristicWithNCCCs"), 500));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for DP1 heuristic without NCCCs");
		testTmp.addTest(new RepeatedTest (new testPreprocessing ("testDP1HeuristicNoNCCCs"), 500));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for DP1 heuristic with NCCCs");
		testTmp.addTest(new RepeatedTest (new testPreprocessing ("testDP1HeuristicWithNCCCs"), 500));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for DP2 heuristic without NCCCs");
		testTmp.addTest(new RepeatedTest (new testPreprocessing ("testDP2HeuristicNoNCCCs"), 500));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for DP2 heuristic with NCCCs");
		testTmp.addTest(new RepeatedTest (new testPreprocessing ("testDP2HeuristicWithNCCCs"), 500));
		testSuite.addTest(testTmp);
		
		return testSuite;
	}
	
	/** 
	 * @see junit.framework.TestCase#setUp()
	 */
	protected void setUp() throws Exception {
		super.setUp();
		
		graph = RandGraphFactory.getRandGraph(maxNbrVars, maxNbrEdges, maxNbrAgents);
		dfs = ch.epfl.lia.frodo.algorithms.dpop.test.UTILpropagationTest.computeDFS(graph);
		
		parser = new XCSPparser<AddableInteger, AddableInteger>(AllTests.generateProblem(graph, false), countNCCCs);
		
//		System.out.println(parser.toDOT());
		variables = new ArrayList<String>(parser.getVariables());
		domains = new HashMap<String, AddableInteger[]>();
		for(String var : variables)
			domains.put(var, parser.getDomain(var));
		counter = 0;
		
		precalculated_h = new HashMap<String, UtilitySolutionSpace<AddableInteger, AddableInteger>>(variables.size());
	}
	
	/** @see junit.framework.TestCase#tearDown() */
	protected void tearDown () throws Exception {
		super.tearDown();
		for(Queue q : queues) {
			q.end();
		}
		this.queues = null;
		precalculated_h = null;
		this.dfs = null;
		this.graph = null;
		this.parser = null;
		this.variables = null;
		this.domains = null;
	}
	
	/** Test for the simple heuristic that only returns zeros, without NCCCs 
	 * @throws Exception 	if an error occurs
	 */
	public void testSimpleHeuristicNoNCCCs() throws Exception {
		testSimpleHeuristic(false);
	}
	
	/** Test for the simple heuristic that only returns zeros, with NCCCs 
	 * @throws Exception 	if an error occurs
	 */
	public void testSimpleHeuristicWithNCCCs() throws Exception {
		testSimpleHeuristic(true);
	}
	
	/**
	 * Test for the simple heuristic that only returns zeros
	 * @author Brammert Ottens, 19 mei 2009
	 * @param countNCCCs 	whether to count NCCCs
	 * @throws Exception 	if an error occurs
	 */
	private void testSimpleHeuristic(boolean countNCCCs) throws Exception {
		testRandom(Preprocessing.SimpleHeuristic.class.getName(), countNCCCs);
	}
	
	/** Test for the DP0 heuristic without NCCCs
	 * @throws Exception 	if an error occurs
	 */
	public void testDP0HeuristicNoNCCCs() throws Exception {
		testDP0Heuristic(false);
	}
	
	/** Test for the DP0 heuristic with NCCCs
	 * @throws Exception 	if an error occurs
	 */
	public void testDP0HeuristicWithNCCCs() throws Exception {
		testDP0Heuristic(true);
	}
	
	/**
	 * Test for the DP0 heuristic
	 * @author Brammert Ottens, 19 mei 2009
	 * @param countNCCCs 	whether to count NCCCs
	 * @throws Exception 	if an error occurs
	 */
	private void testDP0Heuristic(boolean countNCCCs) throws Exception {
		testRandom(Preprocessing.DP0.class.getName(), countNCCCs);
	}
	
	/** Test for the DP1 heuristic, without NCCCs
	 * @throws Exception 	if an error occurs
	 */
	public void testDP1HeuristicNoNCCCs() throws Exception {
		testDP1Heuristic(false);
	}
	
	/** Test for the DP1 heuristic, with NCCCs
	 * @throws Exception 	if an error occurs
	 */
	public void testDP1HeuristicWithNCCCs() throws Exception {
		testDP1Heuristic(true);
	}
	
	/**
	 * Test for the DP1 heuristic
	 * @author Brammert Ottens, 20 mei 2009
	 * @param countNCCCs 	whether to count NCCCs
	 * @throws Exception 	if an error occurs
	 */
	private void testDP1Heuristic(boolean countNCCCs) throws Exception {
		
		// precalculate the heuristics centrally
		for(String variable : dfs.keySet()) {
			if(dfs.get(variable).getParent() == null) {
				calculateDP1(variable, parser.getSolutionSpaces());
			}
		}
		testRandom(Preprocessing.DP1.class.getName(), countNCCCs);
	}
	
	/** Test the heuristics for DP2 without NCCCs
	 * @throws Exception 	if an error occurs
	 */
	public void testDP2HeuristicNoNCCCs() throws Exception {
		this.testDP2Heuristic(false);
	}
	
	/** Test the heuristics for DP2 with NCCCs
	 * @throws Exception 	if an error occurs
	 */
	public void testDP2HeuristicWithNCCCs() throws Exception {
		this.testDP2Heuristic(true);
	}
	
	/**
	 * Test the heuristics for DP2
	 * @author Brammert Ottens, 18 jun 2009
	 * @param countNCCCs 	whether to count NCCCs
	 * @throws Exception 	if an error occurs
	 */
	private void testDP2Heuristic(boolean countNCCCs) throws Exception {
		
		// precalculate the heuristics centrally
		for(String variable : dfs.keySet()) {
			if(dfs.get(variable).getParent() == null) {
				calculateDP2(variable, parser.getSolutionSpaces());
			}
		}
		testRandom(Preprocessing.DP2.class.getName(), countNCCCs);
	}
	
	/**
	 * Tests the Preprocessing listener with \c heuristic
	 * @param heuristic 	the heuristic to use in the test
	 * @param countNCCCs 	whether to count NCCCs
	 * @throws Exception 	if an error occurs
	 */
	private void testRandom(String heuristic, boolean countNCCCs) throws Exception {
		this.countNCCCs = countNCCCs;
		
		int nbrAgents = graph.clusters.size();
		
		// Create the queue network
		queues = new Queue [nbrAgents];
		AllTests.createQueueNetwork(queues, graph, false);

		// Listen for statistics messages
		Queue myQueue = new Queue (false);
		myQueue.addIncomingMessagePolicy(new receiveLowerBounds(heuristic, parser, dfs, countNCCCs));
		QueueIOPipe myPipe = new QueueIOPipe (myQueue);
		for (Queue queue : queues) 
			queue.addOutputPipe(AgentInterface.STATS_MONITOR, myPipe);
		
		// Create the XML parameters
		Element params = new Element ("module");
		params.setAttribute("heuristic", heuristic);
		
		for(String a: parser.getAgents()) {
			Queue queue = queues[Integer.parseInt(a)];

			XCSPparser<AddableInteger, AddableInteger> subProb = parser.getSubProblem(a);
			queue.setProblem(subProb);
			Preprocessing<AddableInteger, AddableInteger> pre = new Preprocessing<AddableInteger, AddableInteger>(subProb, params);
			forwardHeuristicsMessage forwarder = new forwardHeuristicsMessage();
			queue.addIncomingMessagePolicy(pre);
			queue.addIncomingMessagePolicy(forwarder);
		}
		
		// for each variable, create a MessageDFSoutput
		for(String a: parser.getAgents()) {
			Queue queue = queues[Integer.parseInt(a)];
			
			Set<String> variables = parser.getVariables(a);
			for(String var : variables) {
				MessageDFSoutput msg = new MessageDFSoutput(var, dfs.get(var));
				queue.sendMessageToSelf(msg);
			}
		}
		
		// Wait until all messages have been send
		while (true) {
			this.finished_lock.lock();
			try {
				if (counter == variables.size()) { 
					break;
				} else if (! this.finished.await(20, TimeUnit.SECONDS)) {
					fail("Timeout");
				}
			} catch (InterruptedException e) {
				break;
			}
			this.finished_lock.unlock();
		}
		
		myQueue.end();
	}
	
	/**
	 * Centrally calculate the DP1 heuristics
	 * 
	 * @author Brammert Ottens, 20 mei 2009
	 * @param variable The variable for which the heuristics must be calculated
	 * @param spaces A list with all the constraints of the problem
	 * @return A UtilitySolutionSpace containing the heuristic values
	 */
	private UtilitySolutionSpace<AddableInteger, AddableInteger> calculateDP1(String variable, List< ? extends UtilitySolutionSpace<AddableInteger, AddableInteger> > spaces) {
		// find the children of the variable
		List<String> children = dfs.get(variable).getChildren();
		UtilitySolutionSpace<AddableInteger, AddableInteger> h = null;
		
		if(children.size() == 0) {
			String[] vars = new String[1];
			vars[0] = variable;
			AddableInteger[][] domains = new AddableInteger[1][];
			domains[0] = parser.getDomain(variable);
			AddableInteger[] utilities = new AddableInteger[domains[0].length];
			
			for(int i = 0; i < utilities.length; i++) {
				utilities[i] = new AddableInteger(0);
			}
			
			h = new Hypercube<AddableInteger, AddableInteger>(vars, domains, utilities, AddableInteger.PlusInfinity.PLUS_INF);
		} else {
			for(String child : children) {
				UtilitySolutionSpace<AddableInteger, AddableInteger> childH = calculateDP1(child, spaces);
				List<String> childChildren = dfs.get(child).getChildren();
				List<String> pseudo = dfs.get(child).getAllPseudoChildren();
				
				UtilitySolutionSpace<AddableInteger, AddableInteger> localSpace = null;
				
				for (ListIterator< ? extends UtilitySolutionSpace<AddableInteger, AddableInteger> > iterator = spaces.listIterator(); iterator.hasNext(); ) {
					UtilitySolutionSpace<AddableInteger, AddableInteger> space = iterator.next();
					
					// Disregard the space if it does not involve variable or its children
					List<String> constraint_variables = Arrays.asList(space.getVariables());
					if (! constraint_variables.contains(variable) || ! constraint_variables.contains(child)) 
						continue;

					// Disregard the space if it involves any another variable lower than var in the DFS
					boolean isLowestVar = true;
					for (String otherVar : space.getVariables()) {
						if (childChildren.contains(otherVar) || pseudo.contains(otherVar)) {
							isLowestVar = false;
							break;
						}
					}
					
					if(isLowestVar) {
						if(localSpace == null)
							localSpace = space;
						else
							localSpace = localSpace.join(space);
					}
				}

				if(localSpace == null) {
					String[] vars = new String[2];
					vars[0] = variable;
					vars[1] = child;
					AddableInteger[][] domains = new AddableInteger[2][];
					domains[0] = parser.getDomain(variable);
					domains[1] = parser.getDomain(child);
					AddableInteger[] utilities = new AddableInteger[domains[0].length * domains[1].length];

					for(int i = 0; i < utilities.length; i++) {
						utilities[i] = new AddableInteger(0);
					}

					localSpace = new Hypercube<AddableInteger, AddableInteger>(vars, domains, utilities, AddableInteger.PlusInfinity.PLUS_INF);
//				} else {
//					ArrayList<String> constraint_variables = new ArrayList<String>(Arrays.asList(localSpace.getVariables()));
//					constraint_variables.remove(variable);
//					constraint_variables.remove(child);
//					if(constraint_variables.size() > 0)
//						localSpace = localSpace.project(constraint_variables.toArray(new String[0]), false).getSpace();
				}
				
				ArrayList<String> constraint_variables = new ArrayList<String>(Arrays.asList(localSpace.getVariables()));
				constraint_variables.remove(variable);
				if(h == null) {
					h = childH.join(localSpace).blindProject(constraint_variables.toArray(new String[0]), false);
				}
				else
					h = h.join(childH.join(localSpace).blindProject(constraint_variables.toArray(new String[0]), false));
			}
		}
		
		h = h.resolve();
		precalculated_h.put(variable, h);
		
		return h;
	}
	
	/**
	 * Pre calculates the h-values belonging to DP2
	 * 
	 * @author Brammert Ottens, 18 jun 2009
	 * @param variable	The variable for whom to calculate the value
	 * @param spaces	The problem
	 * @return	the h-values for \c variable
	 */
	private UtilitySolutionSpace<AddableInteger, AddableInteger> calculateDP2(String variable, List< ? extends UtilitySolutionSpace<AddableInteger, AddableInteger> > spaces) {
		// find the children of the variable
		List<String> children = dfs.get(variable).getChildren();
		UtilitySolutionSpace<AddableInteger, AddableInteger> h = null;
		
		if(children.size() == 0) {
			String[] vars = new String[1];
			vars[0] = variable;
			AddableInteger[][] domains = new AddableInteger[1][];
			domains[0] = parser.getDomain(variable);
			AddableInteger[] utilities = new AddableInteger[domains[0].length];
			
			for(int i = 0; i < utilities.length; i++) {
				utilities[i] = new AddableInteger(0);
			}
			
			h = new Hypercube<AddableInteger, AddableInteger>(vars, domains, utilities, AddableInteger.PlusInfinity.PLUS_INF);
		} else {
			for(String child : children) {
				UtilitySolutionSpace<AddableInteger, AddableInteger> childH = calculateDP2(child, spaces);
				assert(childH.getVariables().length == 1);
				List<String> childChildren = dfs.get(child).getChildren();
				List<String> pseudoChildren = dfs.get(child).getAllPseudoChildren();
				
				UtilitySolutionSpace<AddableInteger, AddableInteger> localSpace = null;
				UtilitySolutionSpace<AddableInteger, AddableInteger> c = null;
				
				for (ListIterator< ? extends UtilitySolutionSpace<AddableInteger, AddableInteger> > iterator = spaces.listIterator(); iterator.hasNext(); ) {
					UtilitySolutionSpace<AddableInteger, AddableInteger> space = iterator.next();
					
					// Disregard the space if it does not involve the child or on of its ancestors
					ArrayList<String> constraint_variables = new ArrayList<String> (Arrays.asList(space.getVariables()));
					if( ! constraint_variables.contains(child) ) 
						continue;

					// Disregard the space if it involves any another variable lower than var in the DFS
					boolean isLowestVar = true;
					for (String otherVar : space.getVariables()) {
						if (childChildren.contains(otherVar) || pseudoChildren.contains(otherVar)) {
							isLowestVar = false;
							break;
						}
					}
					
					if(isLowestVar) {
						if(constraint_variables.contains(variable)) {
							constraint_variables.remove(variable);
							constraint_variables.remove(child);
							if(constraint_variables.size() == 0)
								c = space;
							else 
								c = space.blindProject(constraint_variables.toArray(new String[0]), false);
							assert c.getVariables().length == 2;
						} else {
							constraint_variables.remove(child);
							if(localSpace == null)
								localSpace = space.blindProject(constraint_variables.toArray(new String[0]), false);
							else
								localSpace = localSpace.join(space.blindProject(constraint_variables.toArray(new String[0]), false));
						}
					}
				}

				if(localSpace == null) {
					String[] vars = new String[1];
					vars[0] = child;
					AddableInteger[][] domains = new AddableInteger[1][];
					domains[0] = parser.getDomain(child);
					AddableInteger[] utilities = new AddableInteger[domains[0].length];

					for(int i = 0; i < utilities.length; i++) {
						utilities[i] = new AddableInteger(0);
					}

					localSpace = new Hypercube<AddableInteger, AddableInteger>(vars, domains, utilities, AddableInteger.PlusInfinity.PLUS_INF);
				}
				
				if(c == null)  {
					String[] vars = new String[1];
					vars[0] = variable;
					AddableInteger[][] domains = new AddableInteger[1][];
					domains[0] = parser.getDomain(variable);
					AddableInteger[] utilities = new AddableInteger[domains[0].length];

					for(int i = 0; i < utilities.length; i++) {
						utilities[i] = new AddableInteger(0);
					}

					c = new Hypercube<AddableInteger, AddableInteger>(vars, domains, utilities, AddableInteger.PlusInfinity.PLUS_INF);
				}
				
				if(h == null) {
					h = (childH.join(localSpace).join(c)).blindProject(child, false);
					assert(h.getVariables().length == 1);
				} else {
					h = h.join(childH.join(localSpace).join(c).blindProject(child, false));
					assert(h.getVariables().length == 1);
				}
			}
		}
		
		assert(h.getVariables().length == 1);
		
		h = h.resolve();
		precalculated_h.put(variable, h);
		
		return h;
	}
	
	/**
	 * 
	 * @author brammert
	 *
	 */
	private class receiveLowerBounds implements IncomingMsgPolicyInterface<String> {

		/** The test queue*/
		Queue queue;
		
		/** The heuristic who's output should be tested */
		String heuristic;
		
		/** The random problem */
		private XCSPparser<AddableInteger, AddableInteger> parser;
		
		/** The DFS belonging to the problem */
		private Map<String, DFSview> dfs;

		/** Whether to test the counting of NCCCs */
		private final boolean countNCCCs;
		
		/**
		 * Constructor
		 * @param heuristic 	The heuristic to be tested
		 * @param parser 		A representation of the global problem
		 * @param dfs 			The DFS structure used for this problem
		 * @param countNCCCs 	whether to test the counting of NCCCs
		 */
		public receiveLowerBounds(String heuristic, XCSPparser<AddableInteger, AddableInteger> parser, Map<String, DFSview> dfs, boolean countNCCCs) {
			this.heuristic = heuristic;
			this.parser = parser;
			this.dfs = dfs;
			this.countNCCCs = countNCCCs;
		}
		
		/** 
		 * 
		 * @see ch.epfl.lia.frodo.communication.IncomingMsgPolicyInterface#getMsgTypes()
		 */
		public Collection<String> getMsgTypes() {
			ArrayList <String> msgTypes = new ArrayList <String> (3);
			msgTypes.add(Preprocessing.HEURISTICS_MSG_TYPE);
			return msgTypes;
		}

		/**
		 * 
		 * @see IncomingMsgPolicyInterface#notifyIn(Message)
		 */
		@SuppressWarnings("unchecked")
		public void notifyIn(Message msg) {
			
			if(msg.getType().equals(Preprocessing.HEURISTICS_MSG_TYPE)) {
				MessageWrapper msgWrap = queue.getCurrentMessageWrapper();
				if (this.countNCCCs) 
					assertTrue (msgWrap.getNCCCs() >= 0);
				
				BoundsMsg<AddableInteger, AddableInteger> castMsg = (BoundsMsg<AddableInteger, AddableInteger>)msg;
				
				String var = castMsg.getSender();
				UtilitySolutionSpace<AddableInteger, AddableInteger> lb = castMsg.getBounds();
				
				if(heuristic.equals(Preprocessing.SimpleHeuristic.class.getName())) {
					for(AddableInteger value : domains.get(var)) {
						AddableInteger[] vals = {value};
						assertEquals(new AddableInteger(0), lb.getUtility(vals));						
					}
				} else if(heuristic.equals(Preprocessing.DP0.class.getName())) {
					
					List< ? extends UtilitySolutionSpace<AddableInteger, AddableInteger> > spaces = parser.getSolutionSpaces();
					
					// for each child, find the local problem and determine
					// the sum of the minimal utilities
					List<String> children = dfs.get(var).getChildren();
					AddableInteger sum = new AddableInteger(0);

					for(String child : children) {
						DFSview view = dfs.get(child);
						
						// set the lower neighbours
						List<String> childsChildren = view.getChildren();
						List<String> pseudo = view.getAllPseudoChildren();
						
						// set the constraints this variable is responsible for
						for (ListIterator< ? extends UtilitySolutionSpace<AddableInteger, AddableInteger> > iterator = spaces.listIterator(); iterator.hasNext(); ) {
							UtilitySolutionSpace<AddableInteger, AddableInteger> space = iterator.next();
							
							// Disregard the space if it does not involve var
							if (! Arrays.asList(space.getVariables()).contains(child)) 
								continue;
							
							// Disregard the space if it involves any another variable lower than var in the DFS
							boolean isLowestVar = true;
							for (String otherVar : space.getVariables()) {
								if (childsChildren.contains(otherVar) || pseudo.contains(otherVar)) {
									isLowestVar = false;
									break;
								}
							}

							if (isLowestVar) { // record the space
								sum = sum.add(space.blindProjectAll(false));
							}
						}
					}
					
					for(AddableInteger value : domains.get(var)) {
						AddableInteger[] vals = {value};
						assertEquals(sum, lb.getUtility(vals));						
					}
				} else if(heuristic.equals(Preprocessing.DP1.class.getName())) {
					assertEquals(precalculated_h.get(var), lb);
				} else if(heuristic.equals(Preprocessing.DP2.class.getName())) {
					assertEquals(precalculated_h.get(var), lb);
				}
				
				finished_lock.lock();
				if (++counter >= variables.size()) 
					finished.signal();
				finished_lock.unlock();
			}
			
		}
		
		/**
		 * 
		 * @see ch.epfl.lia.frodo.communication.IncomingMsgPolicyInterface#setQueue(ch.epfl.lia.frodo.communication.Queue)
		 */
		public void setQueue(Queue queue) {
			this.queue = queue;
			
		}
		
	}
	
	/**
	 * This listener simply forwards all Heuristics messages to the central stats reporter
	 * @author Brammert Ottens, 19 mei 2009
	 *
	 */
	private class forwardHeuristicsMessage implements IncomingMsgPolicyInterface<String> {
	
		/** the queue */
		Queue queue;

		/**
		 * @see ch.epfl.lia.frodo.communication.IncomingMsgPolicyInterface#getMsgTypes()
		 */
		public Collection <String> getMsgTypes() {
			ArrayList<String> msgTypes = new ArrayList<String>(1);
			msgTypes.add(Preprocessing.HEURISTICS_MSG_TYPE);
			return msgTypes;
		}

		/** 
		 * @see IncomingMsgPolicyInterface#notifyIn(Message)
		 */
		@SuppressWarnings("unchecked")
		public void notifyIn(Message msg) {
			
			if(msg.getType().equals(Preprocessing.HEURISTICS_MSG_TYPE)) {
				BoundsMsg<AddableInteger, AddableInteger> msgCast = (BoundsMsg<AddableInteger, AddableInteger>)msg;
				if(msgCast.getReceiver() == null)
					queue.sendMessage(AgentInterface.STATS_MONITOR, msg);
			}
		}
		
		/**
		 * @see ch.epfl.lia.frodo.communication.IncomingMsgPolicyInterface#setQueue(ch.epfl.lia.frodo.communication.Queue)
		 */
		public void setQueue(Queue queue) {
			this.queue = queue;
		}
	}

}
