/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

package ch.epfl.lia.frodo.algorithms.dpop.param.test;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import junit.extensions.RepeatedTest;
import junit.framework.TestSuite;

import org.jdom.Element;

import ch.epfl.lia.frodo.algorithms.RandGraphFactory;
import ch.epfl.lia.frodo.algorithms.dpop.param.ParamUTIL;
import ch.epfl.lia.frodo.algorithms.dpop.param.ParamUTIL.OptUtilMessage;
import ch.epfl.lia.frodo.algorithms.dpop.test.UTILpropagationTest;
import ch.epfl.lia.frodo.communication.Message;
import ch.epfl.lia.frodo.solutionSpaces.Addable;
import ch.epfl.lia.frodo.solutionSpaces.AddableInteger;
import ch.epfl.lia.frodo.solutionSpaces.AddableReal;
import ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace;

/** JUnit test for the class ParamUTIL
 * @author Thomas Leaute
 */
public class ParamUTILtest extends UTILpropagationTest {
	
	/** Maximum number of variables in the random graph 
	 * @note Must be at least 2. 
	 */
	private final int maxNbrVars = 5;
	
	/** Maximum number of edges in the random graph */
	private final int maxNbrEdges = 25;

	/** Maximum number of agents */
	private final int maxNbrAgents = 5;

	/** Constructor 
	 * @param useTCP 		whether to use TCP pipes or shared memory pipes
	 * @param useXML 		whether to use the XML-based constructor
	 * @param utilClass 	the class to use for utility values
	 */
	public ParamUTILtest(boolean useTCP, boolean useXML, Class< ? extends Addable<?> > utilClass) {
		super (useTCP, useXML, utilClass, false, false, false);
	}

	/** @return the test suite for this test */
	static public TestSuite suite () {
		TestSuite testSuite = new TestSuite ("Tests for ParamUTIL");
		
		TestSuite testTmp = new TestSuite ("Tests for the ParamUTIL propagation protocol using shared memory pipes and integer utilities");
		testTmp.addTest(new RepeatedTest (new ParamUTILtest (false, false, AddableInteger.class), 200));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the ParamUTIL propagation protocol with XML support using shared memory pipes and integer utilities");
		testTmp.addTest(new RepeatedTest (new ParamUTILtest (false, true, AddableInteger.class), 200));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the ParamUTIL propagation protocol using TCP pipes and integer utilities");
		testTmp.addTest(new RepeatedTest (new ParamUTILtest (true, false, AddableInteger.class), 100));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the ParamUTIL propagation protocol with XML support using TCP pipes and integer utilities");
		testTmp.addTest(new RepeatedTest (new ParamUTILtest (true, true, AddableInteger.class), 100));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the ParamUTIL propagation protocol using shared memory pipes and real utilities");
		testTmp.addTest(new RepeatedTest (new ParamUTILtest (false, false, AddableReal.class), 200));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the ParamUTIL propagation protocol with XML support using shared memory pipes and real utilities");
		testTmp.addTest(new RepeatedTest (new ParamUTILtest (false, true, AddableReal.class), 200));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the ParamUTIL propagation protocol using TCP pipes and real utilities");
		testTmp.addTest(new RepeatedTest (new ParamUTILtest (true, false, AddableReal.class), 100));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the ParamUTIL propagation protocol with XML support using TCP pipes and real utilities");
		testTmp.addTest(new RepeatedTest (new ParamUTILtest (true, true, AddableReal.class), 100));
		testSuite.addTest(testTmp);
		
		return testSuite;
	}

	/** @see UTILpropagationTest#setUp() */
	@Override
	protected void setUp () {
		super.setUp();
		graph = RandGraphFactory.getRandGraph(maxNbrVars, maxNbrEdges, maxNbrAgents);
		dfs = computeDFS(graph);
	}
	
	/** Creates a new Listener
	 * @param <U> 					the type used for utility values
	 * @param useTCP 				\c true whether TCP pipes should be used instead of QueueIOPipes
	 * @param useXML 				whether we should use the constructor that takes in XML elements or the manual constructor
	 * @param parameters 			the parameters for the module under test
	 * @param countNCCCs 			whether to count NCCCs
	 * @param ignoreHypercubeNCCCs	whether to ignore hypercube NCCCs
	 * @return 						the new listener
	 * @throws Exception 			if an error occurs
	 */
	@Override
	protected < U extends Addable<U> > Listener<U> newListenerInstance(boolean useTCP, boolean useXML, Element parameters, boolean countNCCCs, boolean ignoreHypercubeNCCCs) 
	throws Exception {
		return new Listener<U> (useTCP, useXML, parameters);
	}

	/** The listener that checks the messages sent by the ParamUTIL listeners
	 * @param <U> 	the type used for utility values
	 */
	protected class Listener < U extends Addable<U> > extends UTILpropagationTest.Listener<U> {
		
		/** Optimal utility of each connected component (identified by its root), conditioned on the values of the parameters (if any) */
		private HashMap< String, UtilitySolutionSpace<AddableInteger, U> > optUtils = new HashMap< String, UtilitySolutionSpace<AddableInteger, U> > ();
		
		/** Used to synchronize access to optUtils, which can be \c null */
		private final Object optUtils_lock = new Object ();
		
		/** Constructor that tests the ParamUTIL propagation protocol on a random DFS 
		 * @param useTCP 						\c true whether TCP pipes should be used instead of QueueIOPipes
		 * @param useXML 						whether we should use the constructor that takes in XML elements or the manual constructor
		 * @param parameters 					the parameters for the module under test
		 * @throws Exception 	if an error occurs
		 */
		public Listener (boolean useTCP, boolean useXML, Element parameters) throws Exception {
			super (useTCP, useXML, parameters, ParamUTIL.class, true, false, false);
		}

		/** @see ch.epfl.lia.frodo.algorithms.dpop.test.UTILpropagationTest.Listener#checkOutput() */
		@Override
		protected void checkOutput() {
			
			// Compute optimal utility value of each connected component while checking separators
			HashMap< String, UtilitySolutionSpace<AddableInteger, U> > optUtils2 = new HashMap< String, UtilitySolutionSpace<AddableInteger, U> > ();
			for (String var : graph.nodes) {
				
				// Check if the variable is a root
				if (dfs.get(var).getParent() == null) {
					
					// Simulate the UTIL propagation to compute the optimal utility value for this root's DFS tree
					optUtils2.put(var, simulateUTIL (var));
				}
			}
			
			// Check that the optimal utility values are the same
			assertEquals (optUtils2.size(), optUtils.size());
			for (Map.Entry< String, UtilitySolutionSpace<AddableInteger, U> > entry : optUtils.entrySet()) 
				assertTrue (optUtils2.get(entry.getKey()).equivalent(entry.getValue()));
			
		}

		/** @see ch.epfl.lia.frodo.algorithms.dpop.test.UTILpropagationTest.Listener#getMsgTypes() */
		@Override
		public Collection<String> getMsgTypes() {
			ArrayList<String> types = new ArrayList<String> (2);
			types.add(ParamUTIL.SEPARATOR_MSG_TYPE);
			types.add(ParamUTIL.OPT_PARAM_UTIL_MSG_TYPE);
			return types;
		}

		/** @see ch.epfl.lia.frodo.algorithms.dpop.test.UTILpropagationTest.Listener#notifyIn(Message) */
		@Override
		@SuppressWarnings("unchecked")
		public void notifyIn(Message msg) {
			
			if (msg.getType().equals(ParamUTIL.OPT_PARAM_UTIL_MSG_TYPE)) { // message sent by a root containing the optimal utility value
				OptUtilMessage<AddableInteger, U> msgCast = (OptUtilMessage<AddableInteger, U>) msg;
				synchronized (optUtils_lock) {
					optUtils.put(msgCast.getRoot(), msgCast.getUtility());
				}
			}
			
			super.notifyIn(msg);
		}

	}
	
}
